% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overdispersion.R
\name{gampoi_overdispersion_mle}
\alias{gampoi_overdispersion_mle}
\title{Estimate the Overdispersion for a Vector of Counts}
\usage{
gampoi_overdispersion_mle(
  y,
  mean = base::mean(y),
  model_matrix = matrix(1, nrow = length(y), ncol = 1),
  do_cox_reid_adjustment = TRUE,
  subsample = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{y}{a numeric or integer vector with the counts for which
the overdispersion is estimated}

\item{mean}{a numeric vector of either length 1 or \code{length(y)}
with the predicted value for that sample. Default: \code{mean(y)}.}

\item{model_matrix}{a numeric matrix that specifies the experimental
design. It can be produced using \code{stats::model.matrix()}.
Default: \code{matrix(1, nrow = length(y), ncol = 1)}, which is the
model matrix for a 'just-intercept-model'.}

\item{do_cox_reid_adjustment}{the classical maximum likelihood estimator of the \code{overdisperion} is biased
towards small values. McCarthy \emph{et al.} (2012) showed that it is preferable to optimize the Cox-Reid
adjusted profile likelihood.\cr
\code{do_cox_reid_adjustment} can be either be \code{TRUE} or \code{FALSE} to indicate if the adjustment is
added during the optimization of the \code{overdispersion} parameter. Default: \code{TRUE}.}

\item{subsample}{the estimation of the overdispersion is the slowest step when fitting
a Gamma-Poisson GLM. For datasets with many samples, the estimation can be considerably sped up
without loosing much precision by fitting the overdispersion only on a random subset of the samples.
Default: \code{FALSE} which means that the data is not subsampled. If set to \code{TRUE}, at most 1,000 samples
are considered. Otherwise the parameter just specifies the number of samples that are considered
for each gene to estimate the overdispersion.}

\item{verbose}{a boolean that indicates if information about the individual steps are printed
while fitting the GLM. Default: \code{FALSE}.}
}
\value{
The function returs a list with the following elements:
\describe{
\item{\code{estimate}}{the numerical estimate of the overdispersion.}
\item{\code{iterations}}{the number of iterations it took to calculate
the result.}
\item{\code{method}}{the method that was used to calculate the
overdispersion: either \code{"conventional"} or \code{"bandara"}.}
\item{\code{message}}{additional information about the fitting process.}
}
}
\description{
Estimate the Overdispersion for a Vector of Counts
}
\details{
The function employs a rough heuristic to decide if the iterative
or the Bandara approach is used to calculate the overdispersion. If
\code{max(y) < length(y)} Bandara's approach is used, otherwise the
conventional one is used.
}
\examples{
 set.seed(1)
 # true overdispersion = 2.4
 y <- rnbinom(n = 10, mu = 3, size = 1/2.4)
 # estimate = 1.7
 gampoi_overdispersion_mle(y)


 # true overdispersion = 0
 y <- rpois(n = 10, lambda = 3)
 # estimate = 0
 gampoi_overdispersion_mle(y)
 # with different mu, overdispersion estimate changes
 gampoi_overdispersion_mle(y, mean = 15)
 # Cox-Reid adjustment changes the result
 gampoi_overdispersion_mle(y, mean = 15, do_cox_reid_adjustment = FALSE)


 # Many very small counts, true overdispersion = 50
 y <- rnbinom(n = 1000, mu = 0.01, size = 1/50)
 summary(y)
 # estimate = 31
 gampoi_overdispersion_mle(y)

}
\seealso{
\code{\link[=glm_gp]{glm_gp()}}
}
